"""
Check the requirements.
    TODO list:
        remove plus suffixes in get_installed
"""

import re, subprocess, os
try:
    from pip_dater.util import runtime_output
    from pip_dater.database import DB
except:
    from util import runtime_output
    from database import DB


def change_version(req, name, ver_new):
    # 'flags=re.MULTILINE' is required for ^ to match the begin of line
    return re.sub(f"(?m)^{name}==.+", f"{name}=={ver_new}", req)

def get_installed(cleanup=True, remove_plus=False):
    """Get the list of requirements.
    Optionally removed references to packages installed locally from wheels."""
    req = subprocess.run(['pip', 'freeze'], capture_output=True, text=True).stdout
    if cleanup:
        # local files are marked with '@': fix them
        req = re.sub(r'(.+) @ .+-([0-9.]+)[-\.].*', r'\1==\2', req)
    if remove_plus:
        pass  # TODO: remove plus suffixes
    return req

def check(req, verbose=False, filter_errors=False):
    """Verify a list of requirements."""
    temp = '_check_requirements.txt'
    with open(temp, 'w') as fw:
        fw.write(req)
    err, out = runtime_output(f"pip install --dry-run --ignore-installed -r {temp}",
                              verbose=verbose)
    os.remove(temp)
    if err and filter_errors:
        # keep only the interesting lines of the log
        # bad_lines = re.findall(r'(?ms)^ERROR: (?!ResolutionImpossible).+?\n|'
        #                        r'(?<=The conflict is caused by:\n).+?(?=\n\n)',
        #                        out)
        bad_lines = re.findall(r'(?ms)(?<=The conflict is caused by:\n).+?(?=\n\n)',
                               out)
        out = '\n'.join(bad_lines)
    return err, out

def check_outdated(req_installed, candidate, probably_good=[], verbose=False):
    """Verify a list of requirements against proposed upgrades."""
    db = DB()
    good_candidate = []
    if probably_good:
        ottimismo = [x for x in candidate if x[0] in probably_good]
        req = req_installed
        for name, _, ver_new in ottimismo:
            req = change_version(req, name, ver_new)
        print(f"  Group testing: {probably_good}")
        err, out = check(req, verbose=verbose)
        if err:
            print('    check FAILED.')
        else:
            print('    check passed.')
            good_candidate = probably_good
            candidate = [x for x in candidate if x[0] not in probably_good]
    for name, _, ver_new in candidate:
        print(f"  Testing: {name} {ver_new}")
        fullname = f"{name}=={ver_new}"
        if fullname in db:
            blockers = db[fullname]
            blockers_found = [b for b in blockers if b in req_installed]
            found_in_db = len(blockers_found) > 0
        else:
            found_in_db = False
        if found_in_db:
            err = True
            out = f"    Blockers found in DB: {','.join(blockers_found)}"
        else:
            req = change_version(req_installed, name, ver_new)
            err, out = check(req, verbose=verbose, filter_errors=True)
        if err:
            db_str = ' (blockers found in DB)' if found_in_db else ''
            print(f'    check FAILED{db_str}.')
        else:
            print('    check passed.')
            good_candidate.append(name)
        # update db in case of fail
        if err and fullname not in db:
            blockers = re.findall(r'(\w.+?)(\[.+\])? (.+) depends on ', out)
            bb = [f"{n}=={v}" for n, _, v in blockers]
            db[fullname] = bb
    return good_candidate

def probably_good_db(outdated, req_installed, verbose=False):
    """Verify if the proposed are compatible with the installed using the DB.
    Return a list of candidates."""
    db = DB()
    candidate = set(name for name, _, _ in outdated)
    for name, ver_old, ver_new in outdated:
        print(f"  Testing: {name} {ver_new}")
        fullname = f"{name}=={ver_new}"
        if fullname in db:
            # block the package if at least a blocker is installed
            blockers = db[fullname]
            blockers_found = [b for b in blockers if b in req_installed]
            if blockers_found:
                if verbose:
                    print(f"    Blockers found: {','.join(blockers_found)}")
                candidate.discard(name)
    return sorted(list(candidate))

##########################
if __name__ == "__main__":
    # # test output filter
    # out = """    python-lsp-server[all] 1.14.0 depends on autopep8<2.1.0 and >=2.0.4; extra == "all"'"""
    # blockers = re.findall(r'(\w.+?)(\[.+\])? (.+) depends on ', out)
    # print(blockers)

    # test check_outdated
    print("Testing upgradable packages")
    req_installed = get_installed()
    outdated = [
                # ['scipy', '1.16.3', '1.17.0'],
                # ['black', '25.12.0', '26.1.0'],
                # ['async-lru', '2.0.5', '2.1.0'],
                # ['ruff', '0.14.10', '0.14.13'],
                ['autopep8', '2.0.4', '2.3.2'],
                ['pyflakes', '3.2.0', '3.4.0'],
                ['tinycss2', '1.4.0', '1.5.1']
                ]
    print("Finding 'probably good' upgrades.")
    probably_good = probably_good_db(outdated, req_installed, verbose=True)
    # probably_good = []
    print(probably_good)
    # good_candidate = check_outdated(req_installed, outdated,
    #                                 probably_good=probably_good, verbose=False)
    # print('Good:', good_candidate)

    # # test check
    # print('Checking current installation')
    # req = get_installed()
    # # req = req.replace('autopep8==2.0.4', 'autopep8==2.3.2')
    # # req = req.replace('tinycss2==1.4.0', 'tinycss2==1.5.1')
    # err, out = check(req)
    # if err: print('   ', out)



"""
Helper functions.
"""
import subprocess


def runtime_output(args, indent='  ', verbose=False):
    """Display in realtime the output of the subprocess.
    Return the error code and the whole output (stdout+stderr).
    """
    cursor = '-\\|/'
    output = ''
    with subprocess.Popen(args, shell=False, text=True, bufsize=1,                     
                          stdout=subprocess.PIPE, stderr=subprocess.STDOUT
                          ) as process:
        for line in process.stdout:
            output += line
            if verbose:
                print(indent, line, end='')
            else:
                print(indent, cursor[0], '\r', end='')
                cursor = cursor[1:] + cursor[0]
        if not verbose: print(indent, ' ', '\r', end='')  # cleanup
        process.communicate()
        process.wait()  # DEBUG: is this really required?
        returncode = process.returncode
    return returncode, output


##########################
if __name__ == "__main__":
    # test runtime_output
    err1, out1 = runtime_output("pip install --help", verbose=True)
    err2, out2 = runtime_output("pip install --help", verbose=False)
    print(err1, len(out1))
    print(err2, len(out2))



"""
Main.
TODO list:
    main
        option to reset DB
    check/upgrade:
        remove plus suffixes in get_installed
    download:
        add option to keep existing files in wheelhouse
"""
__version__ = "1.1.2"
__dev_mode__ = False


import subprocess, argparse
try:
    from pip_dater.download import download
    from pip_dater.check import get_installed, check, check_outdated, probably_good_db
except:
    from download import download
    from check import get_installed, check, check_outdated, probably_good_db


def main(dev_stdout='', cmdline='', skip_download=False):
    parser = argparse.ArgumentParser(prog="pip-dater",
        description="A maintenance tool for the python packages.",
        formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("cmd", choices=["check", "upgrade", "download"],
                         help="check: perform integrity check on installed packages.\n"
                                "upgrade: upgrade the packages.\n"
                                "download: download the wheels for offline installation.")
    parser.add_argument("-v", "--version", action="version", version=__version__)
    parser.add_argument("-e", "--exclude", nargs="+", help="[upgrade] Exclude packages you don't want to upgrade")
    parser.add_argument("-y", "--yes", action="store_true", help="Automatically execute (no user prompt))")
    parser.add_argument("-n", "--no", action="store_true", help="Don't execute (no user prompt))")
    parser.add_argument("--archive", action="store_true", help="[download] Compress the wheelhouse")
    parser.add_argument("-i", "--include-dir", nargs=1, help="[download] Add all the files from the specified folder.")
    if cmdline:
        args = parser.parse_args(cmdline.split())
    else:
        args = parser.parse_args()

    # Common step: get installed packages/requirements
    req_installed = get_installed()
    print(f"Installed packages: {len(req_installed.splitlines())}")

    if args.cmd == 'check':
        print("Performing integrity test on installed packages")
        err, out = check(req_installed)
        if err:
            print('   Check failed! Log:\n', out)
        else:
            print('   Check passed.')
        return
    elif args.cmd == 'download':
        download(yes=args.yes, archive=args.archive, skip_download=skip_download,
                 include=args.include_dir[0] if args.include_dir else None)
        return
    elif args.cmd not in ('upgrade'):
        raise Exception('Unknown command!')


    ### CMD = UPGRADE (from here to end)
    print("Getting outdated packages")
    if dev_stdout:
        text = dev_stdout
    else:
        text = subprocess.run(['pip', 'list', '--outdated'],
                              capture_output=True, text=True).stdout
    outdated = [l.split()[:3] for l in text.splitlines()[2:]]  # skip header
    print(f"   Found {len(outdated)} outdated packages")
    print(f"   {[x[0] for x in outdated]}")
    excluded = args.exclude or []  # replace None
    print('Excluded from upgrade:', excluded)
    candidate = [x for x in outdated if x[0] not in excluded]
    print("Finding 'probably good' upgrades.")
    probably_good = probably_good_db(candidate, req_installed, verbose=True)
    print("Testing upgradable packages")
    good_candidate = check_outdated(req_installed, candidate, probably_good=probably_good)
    print(f"Upgradable packages: {len(good_candidate)}")
    if not good_candidate:
        print('Nothing to upgrade. DONE.')
        return

    print('To be upgraded:', good_candidate)
    if args.no:
        ans = 'n'
        print('   Upgrade skipped (--no flag in command line)')
    elif args.yes:
        ans = 'y'
        print('   Upgrade forced (--yes flag in command line)')
    else:
        ans = input('  Proceed with upgrade? (y/n) ')
    if ans.lower() in ('y', 'yes'):
        subprocess.run(['python', '-m', 'pip', 'install', '--upgrade'] + good_candidate)


############
if __name__ == "__main__":
    if not __dev_mode__:
        main()
    else:
        # DEVELOPMENT (remember to set __dev_mode__)
        # main(cmdline='-h')
        # main(cmdline='download -y --archive --include-dir test_include', skip_download=True)
        # main(cmdline='check')
        main(cmdline='upgrade -n -e pyinstaller selenium',
             dev_stdout="""Package     Version Latest Type
            ----------- ------- ------ -----
            autopep8    2.0.4   2.3.2  wheel
            debugpy     1.8.19  1.8.20 wheel
            flake8      7.1.2   7.3.0  wheel
            ipykernel   6.31.0  7.1.0  wheel
            jupyterlab  4.5.2   4.5.3  wheel
            nbconvert   7.16.6  7.17.0 wheel
            notebook    7.5.2   7.5.3  wheel
            psutil      7.2.1   7.2.2  wheel
            pycodestyle 2.12.1  2.14.0 wheel
            pycparser   2.23    3.0    wheel
            pyflakes    3.2.0   3.4.0  wheel
            pyinstaller 6.17.0  6.18.0 wheel
            pytokens    0.4.0   0.4.1  wheel
            QtAwesome   1.4.0   1.4.1  wheel
            scipy       1.16.3  1.17.0 wheel
            selenium    4.38.0  4.40.0 wheel
            tinycss2    1.4.0   1.5.1  wheel
            wcwidth     0.5.0   0.5.2  wheel""")



"""
Download.
TODO:
    add option to keep existing files in wheelhouse.
"""
import shutil, os, re, zipfile
from datetime import date
try:
    from pip_dater.util import runtime_output
    from pip_dater.check import get_installed
except:
    from util import runtime_output
    from check import get_installed


def download(yes=False, archive=False, include=None, skip_download=False):
    """Download all the packages and create a script for offline installation."""
    print('Getting and cleaning requirements')
    req_installed = get_installed()
    if yes:
       ans = 'y'
       print('Rebuild wheelhouse forced (--yes flag)')
    else:
       ans = input('Remove and rebuild wheelhouse? (y/n) ')
    if ans.lower() in ('y', 'yes'):
        shutil.rmtree('wheelhouse', ignore_errors=True)
        os.mkdir('wheelhouse')
        with open('wheelhouse\\_packages.txt', 'w') as fw:
            fw.write(req_installed)
        if include:
            print(f"Copying the packages from the specified folder: {include}")
            for filename in os.listdir(include):
                print(f"  {filename}")
                shutil.copy(f"{include}\\{filename}", 'wheelhouse')
        print('Downloading base packages (pip/wheel/setuptools)')
        runtime_output('pip download -d wheelhouse pip wheel setuptools', verbose=True)
        print('Downloading everything else ...')
        if not skip_download:
            runtime_output('pip download -r wheelhouse\\_packages.txt -d wheelhouse',
                           verbose=True)
        # create the install batch file, using the exact file names
        print('\nCreate the install batch file')
        pip        = [x for x in os.listdir('wheelhouse') if re.match(r'pip-\d+', x)][-1]
        wheel      = [x for x in os.listdir('wheelhouse') if re.match(r'wheel-\d+', x)][-1]
        setuptools = [x for x in os.listdir('wheelhouse') if re.match(r'setuptools-\d+', x)][-1]
        txt = (
            "@echo off\n\n"
            "echo **** STEP1: update PIP/SETUPTOOLS/WHEEL ****\n"
            f"python -m pip install wheelhouse\\{pip}\n"
            f"pip install wheelhouse\\{setuptools}\n"
            f"pip install wheelhouse\\{wheel}\n\n"
            "echo **** STEP2: install the libraries ****\n"
            "pip install --no-index --find-links wheelhouse -r wheelhouse\\_packages.txt\n\n"
            "echo **** ALL DONE ****\npause\n")
        with open('offline_install.bat', 'w') as fw:
            fw.write(txt)

        # archive
        if archive:
            print('Archive the wheelhouse folder')
            archive_name = f"wheelhouse_{date.today()}.zip"
            try:
                os.remove(archive_name)
            except:
                pass
            with zipfile.ZipFile(archive_name, 'w', compression=zipfile.ZIP_DEFLATED, compresslevel=6) as zf:
                zf.write('offline_install.bat')
                cursor = '-\\|/'
                for path, _, files in os.walk('wheelhouse'):
                    for fn in files:
                        src_filename = os.path.join(path, fn)
                        zf.write(src_filename)
                        print('   ', cursor[0], '\r', end='')
                        cursor = cursor[1:] + cursor[0]
                print('    ', '\r')

            

##########################
if __name__ == "__main__":
    download(yes=True, archive=True, include='test_include', skip_download=True)



"""
Database of blocking packages:
    entry = 'package==version'
    key = entry, value = list of entries
Every operation is immediately stored in file.
"""
import os
from collections import UserDict


class DB(UserDict):
    def __init__(self):
        """Initialize from file (create an empty file if missing)."""
        super().__init__()
        self.filename = os.path.join(os.environ.get('USERPROFILE'),
                                     '.pip-dater.db')
        try:
            with open(self.filename, 'r') as f:
                raw = [l.split(';') for l in f.read().splitlines()]
        except:
            with open(self.filename, 'w') as fw:
                fw.write('')
            raw = []                
        for l in raw:
            self[l[0]] = l[1].split(',')
        
    def write(self):
        with open(self.filename, 'w') as fw:
            for package, blockers in self.items():
                fw.write(f"{package};{','.join(blockers)}\n")

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self.write()

    def remove(self, key):
        if key in self:
            self.pop(key)
            self.write()


##########################
if __name__ == "__main__":
    db = DB()
    db.remove('test_x')
    db['test_a'] = ['a1']
    db['test_b'] = ['b1', 'b2']
    db['test_c'] = ['c1', 'c2', 'c3']
    db.remove('test_b')
    print(db)
    db = DB()
    print(db)
    db.remove('test_a')
    db.remove('test_c')
       